#include "SignalAcquisition.h"

// Global variables:
SignalAcquisition<float> gSignal(Serial);
bool gMute = true;

void setup()
{
  Serial.begin(115200); // but you'd be well advised to use a microcontroller
                        // that ignores the baud rate parameter and simply
                        // communicates at USB rates
  
  // You can hard-code the signal acquisition parameters here:
  gSignal.samplesPerSecond = 1000.0;
  gSignal.samplesPerBlock = 50;
  gSignal.sourcePins = "22 23"; // or whatever
  gSignal.configure();
  // ...or you can set them up later via serial port commands, thanks
  // to the Keyhole library and the .exposeVariables() method.
}

void loop()
{
  unsigned int now = micros();
  if(!gMute) gSignal.acquire(now);
  
  // That's really all there is to it, but if you want to configure it,
  // control it, and diagnose its timing performance all via the serial
  // port, here's how you integrate with the Keyhole library:
  
  KEYHOLE keyhole(Serial);  
  if(keyhole.begin(now))
  {
    gSignal.exposeVariables(keyhole);    // This exposes "samplesPerSecond", "samplesPerBlock" and "sourcePins" through the Keyhole,
                                         // as well as a couple of variables for timing diagnostics. To enable sampling, set
                                         // samplesPerSecond and samplesPerBlock to positive non-zero values, with at least one pin
                                         // address in sourcePins.
                                         
    if( keyhole.variable("mute", gMute)) // - so "mute=0\n" and "mute=1\n" can be used as start and stop commands.
      gSignal.reset(now);                // Call reset() whenever acquisition is stopped or restarted.
    
    keyhole.end();
  }
}

/*

This example acquires float32 signals from two pins
(although the number of pins and their addresses can be changed
on-the-fly by assigning to "sourcePins" with a serial-port command,
thanks to the Keyhole library). These pins get analogRead() 1000
times per second at regular intervals. Batches of 50 samples are
accumulated in memory and then sent over the serial port using a
simple binary protocol: each block begins with either \x01\x00 to
indicate little-endian or \x00\x01 to indicate big-endian, followed
by a \r\n line-ending, and then the raw packed floats are sent
(readings from all pins on the first sample of the block, then
readings from all pins on the second sample, and so on) followed by
a final \r\n line-ending.

To change the numeric type of the data, you can specify, say,
SignalAcquisition<int> instead of <float> when you create the object.
To make the class do something other than just analogRead() from
the specified pins, you would have to create your own subclass of
SignalAcqusition<float> and overshadow the readSample() method in
your subclass implementation (consult SignalAcquisition.cpp to see
what it currently does).

To debug timing: send  commOverheadMicros=1  via the Serial Monitor,
then  mute=0  to start acquisition, then  mute=1;?  to stop it and
examine variables. Repeat the last two commands a few times. Make
sure idleLoops is reliably > 0; if not, your sketch may be unable
to keep up with the desired sample rate: reduce samplesPerSecond
and/or optimize your loop() implementation. Also verify that
commOverheadMicros remains less than the duration of one sample;
if not, that shows that the block is taking too long to send,
and your samples will be unevenly sampled in time. To fix that,
reduce samplesPerSecond, and/or reduce samplesPerBlock, and/or
reduce the number of sourcePins.

*/