/*
# $BEGIN_SIGNALACQUISITION_LICENSE$
# 
# This file is part of the SignalAcquisition project, a library for
# the Arduino IDE that samples signals at regular intervals and
# sends the resulting samples in batches over serial port using a
# simple binary protocol. 
# 
# Copyright (c) 2023- Jeremy Hill
# 
# Behold the 3-Clause BSD License under which it is hereby released:
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $END_SIGNALACQUISITION_LICENSE$


The following example acquires float32 signals from two pins
(although the number of pins and their addresses can be changed
on-the-fly by assigning to "sourcePins" with a serial-port command,
thanks to the Keyhole library). These pins get analogRead() 1000
times per second at regular intervals. Batches of 50 samples are
accumulated in memory and then sent over the serial port using a
simple binary protocol: each block begins with either \x01\x00 to
indicate little-endian or \x00\x01 to indicate big-endian, followed
by a \r\n line-ending, and then the raw packed floats are sent
(readings from all pins on the first sample of the block, then
readings from all pins on the second sample, and so on) followed by
a final \r\n line-ending::

    #include "SignalAcquisition.h"
  
    // Global variables:
    SignalAcquisition<float> gSignal(Serial);
    bool gMute = true;
  
    void setup()
    {
      Serial.begin(115200); // but you'd be well advised to use a microcontroller
                            // that ignores the baud rate parameter and simply
                            // communicates at USB rates
      
      // You can hard-code the signal acquisition parameters here:
      gSignal.samplesPerSecond = 1000.0;
      gSignal.samplesPerBlock = 50;
      gSignal.sourcePins = "22 23"; // or whatever
      gSignal.configure();
      // ...or you can set them up later via serial port commands, thanks
      // to the Keyhole library and the .exposeVariables() method.
    }
  
    void loop()
    {
      unsigned int now = micros();
      if(!gMute) gSignal.acquire(now);
      
      // That's really all there is to it, but if you want to configure it,
      // control it, and diagnose its timing performance all via the serial
      // port, here's how you integrate with the Keyhole library:
      
      KEYHOLE keyhole(Serial);  
      if(keyhole.begin(now))
      {
        gSignal.exposeVariables(keyhole);    // This exposes "samplesPerSecond", "samplesPerBlock" and "sourcePins" through the Keyhole,
                                             // as well as a couple of variables for timing diagnostics. To enable sampling, set
                                             // samplesPerSecond and samplesPerBlock to positive non-zero values, with at least one pin
                                             // address in sourcePins.
                                             
        if( keyhole.variable("mute", gMute)) // - so "mute=0\n" and "mute=1\n" can be used as start and stop commands.
          gSignal.reset(now);                // Call reset() whenever acquisition is stopped or restarted.
        
        keyhole.end();
      }
    }

To change the numeric type of the data, you can specify, say,
SignalAcquisition<int> instead of <float> when you create the object.
To make the class do something other than just analogRead() from
the specified pins, you would have to create your own subclass of
SignalAcqusition<float> and overshadow the readSample() method in
your subclass implementation (consult SignalAcquisition.cpp to see
what it currently does).

To debug timing: send  commOverheadMicros=1  via the Serial Monitor,
then  mute=0  to start acquisition, then  mute=1;?  to stop it and
examine variables. Repeat the last two commands a few times. Make
sure idleLoops is reliably > 0; if not, your sketch may be unable
to keep up with the desired sample rate: reduce samplesPerSecond
and/or optimize your loop() implementation. Also verify that
commOverheadMicros remains less than the duration of one sample;
if not, that shows that the block is taking too long to send,
and your samples will be unevenly sampled in time. To fix that,
reduce samplesPerSecond, and/or reduce samplesPerBlock, and/or
reduce the number of sourcePins.

*/

#ifndef    __SignalAcquisition_H__
#define    __SignalAcquisition_H__

#include "Keyhole.h" // install the Keyhole library using the library manager of your Arduino IDE

template< typename NUMERICTYPE >
class SignalAcquisition
{
  public:
    SignalAcquisition( Stream & stream=Serial, unsigned int maxPins=16 );
    ~SignalAcquisition();

    // Call exposeVariables(keyhole) inside your `if( keyhole.begin() )` block, making sure you pass that same Keyhole instance. This will make "samplesPerSecond", "samplesPerBlock", "sourcePins" available as Keyhole variables, and will call configure() automatically if they are changed via the Keyhole.
    void exposeVariables( Keyhole & keyhole );

    // configure() is called by exposeVariables() when any variables change via the Keyhole (NB: the Keyhole instance is only needed for composing/sending error messages).
    void configure( Keyhole & keyhole, const String & reportBadValues="*" );
    // If you're not controlling the SignalAcquisition variables via Keyhole, call configure() yourself when any configuration variables change.
    void configure( bool reportBadValues=false ) { Keyhole k( this->stream ); configure( k, reportBadValues ? "*" : "" ); }

    // Call reset() whenever signal acquisition is interrupted or resumed.
    void reset( unsigned int currentTimeInMicros );
    // Call reset() whenever signal acquisition is interrupted or resumed.
    void reset( void ) { reset( micros() ); }

    // Call acquire() in every loop().
    void acquire( unsigned int currentTimeInMicros );
    // Call acquire() in every loop().
    void acquire( void )  { acquire( micros() ); }

    Stream &      stream;             // a reference to the Stream (e.g. Serial) used signal output
    double        samplesPerSecond;   // call configure() after changing samplesPerSecond, samplesPerBlock or sourcePins (this will happen automatically if they're changed via Keyhole, using exposeVariables())
    unsigned int  samplesPerBlock;    // call configure() after changing samplesPerSecond, samplesPerBlock or sourcePins (this will happen automatically if they're changed via Keyhole, using exposeVariables())
    String        sourcePins;         // a space-delimited list of ascii-decimal pin addresses;  call configure() after changing samplesPerSecond, samplesPerBlock or sourcePins (this will happen automatically if they're changed via Keyhole, using exposeVariables())

    unsigned int  commOverheadMicros; // for debugging; set to zero to disable measurement; set to non-zero, and it will be updated each time a block is sent. Check that it is reliably less than the sample period in (if not: sending each block over the serial port is taking too long, and your samples will end up being unevenly sampled in time). This variable exposed through the Keyhole if you use exposeVariables().
    unsigned int  idleLoops;          // for debugging; check that this stays reliably > 0 during acquisition (if not: your sketch is unable to keep up with the desired sample rate). This variable is exposed through the Keyhole (as a read-only variable) if you use exposeVariables().

  protected:
    // readSample() is called by acquire() to read one sample value from each pin. It has been made a separate method so that it is easy to overshadow if you want to make your own SignalAcquisition subclass.
    void readSample( void );

  private:
    unsigned int  mIdleLoopsSoFar;

    unsigned int  mSampleTimeInMicros;
    unsigned int  mSamplePeriodInMicros;

    unsigned int  mMaxPins;
    int *         mSourcePinAddresses;

    NUMERICTYPE * mBufferStart;
    NUMERICTYPE * mBufferCursor;
    NUMERICTYPE * mBufferFull;
    size_t        mBufferSize;
};

// With template classes, you get "undefined reference" errors from the linker unless you either
// (1) in the .h file, include the actual definitions (.cpp content) or (2) in the .cpp file,
// declare `template class SignalAcquisition<X>;` for each allowable X.  Solution (2) makes for
// faster compilation but may make the compiled sketch a few bytes bigger on some boards (e.g.,
// +64 bytes on ItsyBitsy M4, or +72 on Raspberry Pi Pico) while making no difference on some
// other boards (Teensy 3.1, Arduino UNO). See https://stackoverflow.com/a/8752879
// The library implements solution (2) for float, double, uint16_t, uint32_t, int16_t and uint16_t.
// Uncomment the #include line below to additionally implement solution (1) and thereby allow your
// sketch to have a completely open choice of numeric type X.
//#include "SignalAcquisition.cpp"


#endif // __SignalAcquisition_H__
